indexing
	description: "Objects that print static program point information in .decls format %
		%and produce instructions to print program point state in .dtrace format"
	author: "Nadia Polikarpova"
	date: "$Date$"
	revision: "$Revision$"

class
	CI_DAIKON_PROGRAM_POINT_PRINTER

inherit
	CI_PROGRAM_POINT_PRINTER

	CI_OPERATOR_IDENTIFIERS
		export {NONE}
			all
		end

	CI_PRINTABLE_CLASSES

feature -- Processing
	process (a_program_point: CI_PROGRAM_POINT)
			-- Print static information about `a_program_point' and build instructions that print its dynamic state
		local
			ppt_name: STRING
		do
			current_program_point := a_program_point
			create state_printing_instructions.make
			create extra_locals.make
			create {LINKED_LIST [ET_COMPOUND]} alternative_compounds.make

			-- Guard printing instructions if monitoring control is enabled
			if Settings.monitoring_control_enabled then
				current_guard := expression_guard (Settings.Monitoring_control_flag_name, False)
				state_printing_instructions.force_first (current_guard)
			end

			-- Process common program point information
			ppt_name := program_point_name (current_program_point)
			output_ppt_info (ppt_name)
			-- Build printing instructions
			current_compound.force_first (new_line_instruction)
			current_compound.force_first (string_printing_instruction (ppt_name))

			-- Process variables
			from
				current_program_point.unfolded_variables.start
			until
				current_program_point.unfolded_variables.after
			loop
				process_unfolded_variable (current_program_point.unfolded_variables.item)
				current_program_point.unfolded_variables.forth
			end

			-- Output delimiter between program point declarations
			static_output.new_line

			-- Add delimiter output instruction
			insert_instruction_to_back (current_compound, new_line_instruction)
		end

feature -- Output
	program_point_name (a_program_point: CI_PROGRAM_POINT): STRING is
			-- Program point name in Daikon format
		local
			routine_ppt: CI_ROUTINE_POINT
			routine_body_ppt: CI_ROUTINE_BODY_POINT
		do
			Result := a_program_point.surrounding_class.upper_name.twin
			routine_ppt ?= a_program_point
			if routine_ppt /= Void then
					-- `a_program_point' is within routine, routine name should be printed
				Result.append_character (class_feature_separator)
				Result.append (identifier_name (routine_ppt.surrounding_routine.name).lower_name)
				Result.append ("()")
			end
			routine_body_ppt ?= a_program_point
			if routine_body_ppt /= Void then
				Result.append_character (position_separator)
				Result.append (routine_body_ppt.position.out)
			end
			Result.append (kind_separator)
			Result.append (a_program_point.kind (old_suffixes))
		end

	output_caption is
			-- Ouput general information about declarations into static output
		do
		end

	input_program_point_name (a_name: STRING): STRING is
			-- Version of `a_name' in trace file, if `a_name' is a program point name from inferred assertions file; identity otherwise
		local
			start: INTEGER
			kind_suffix: STRING
		do
			Result := a_name.string
			start := a_name.substring_index (kind_separator, 1)
			if start > 0 then
				kind_suffix := a_name.substring (start + kind_separator.count, a_name.count)
				if kind_suffix.is_equal (old_suffixes.aggregative_exit_kind) then
					Result.append (old_suffixes.default_exit_id)
				end
			end
		end

feature {NONE} -- Output
	output_ppt_info (ppt_name: STRING) is
			-- Print the prologue of a new program point declaration
		do
			static_output.put_string (declaration_keyword)
			static_output.new_line
			static_output.put_string (ppt_name)
			static_output.new_line
		end

	output_variable_info (a_variable: CI_VARIABLE) is
			-- Output static information about `a_variable' in Daikon format:
			-- its name, declared type, representation type and comparability information
		require
			a_variable_exists: a_variable /= Void
		do
			static_output.put_string (a_variable.name)
			static_output.put_new_line
			if a_variable.type.kind.is_printable then
				-- Daikon deals better with Java types (e.g. boolean for splitting)
				static_output.put_string (a_variable.type.representation_type)
			else
				static_output.put_string (a_variable.type.base_class.upper_name)
			end
			if a_variable.is_param then
				static_output.put_string (" # isParam=true")
			end
			static_output.put_new_line
			static_output.put_string (a_variable.type.representation_type)
			if a_variable.constant /= Void then
				static_output.put_string (" = ")
				static_output.put_string (constant_value (a_variable.constant))
			end
			static_output.put_new_line
			static_output.put_string (a_variable.comparability.out)
			static_output.put_new_line
		end

	constant_value (a_constant: ET_CONSTANT): STRING is
			-- Value of `a_constant' in Daikon format
		require
			a_constant_exists: a_constant /= Void
		local
			t: ET_TRUE_CONSTANT
			f: ET_FALSE_CONSTANT
			char: ET_CHARACTER_CONSTANT
			int: ET_INTEGER_CONSTANT
			real: ET_REAL_CONSTANT
			str: ET_MANIFEST_STRING
			literal: STRING
			i: INTEGER
		do
			t ?= a_constant
			if t /= Void then
				Result := "1"
			else
				f ?= a_constant
				if f /= Void then
					Result := "0"
				else
					char ?= a_constant
					if char /= Void then
						Result := char.value.code.out
					else
						int ?= a_constant
						if int /= Void then
							literal := int.literal.string
							literal.replace_substring_all ("_", "")
							if int.is_negative then
								literal.prepend ("-")
							end
							Result := literal
						else
							real ?= a_constant
							if real /= Void then
								literal := real.literal.string
								literal.replace_substring_all ("_", "")
								Result := literal
								if real.is_negative then
									literal.prepend ("-")
								end
							else
								str ?= a_constant
								if str /= Void then
									Result := str.value.string
									from
										i := escape_sequences.lower
									until
										i > escape_sequences.upper
									loop
										Result.replace_substring_all (escape_sequences.item (i).value,
											double_escape_sequences.item (i).value)
										i := i + 1
									end
									Result := "%"" + Result + "%""
								end
							end
						end
					end
				end
			end
		ensure
			result_not_empty: not Result.is_empty
		end

feature -- Constants
	nonsensical_keyword_name: STRING is "nonsensical"
			-- Name of Daikon "nonsensical" keyword	

feature {NONE} -- Constants
	class_feature_separator: CHARACTER is '.'
			-- Separator between class and feature name in program point name

	position_separator: CHARACTER is '@'
			-- Separator before position in program point name

	kind_separator: STRING is ":::"
			-- Separator before program point kind in its name

	declaration_keyword: STRING is "DECLARE"
			-- Daikon keyword to declare a program point	

	nonsensical_keyword: ET_REGULAR_MANIFEST_STRING is
			-- Daikon keyword that is used instead variable value when it can't be evaluated
		once
			create Result.make (nonsensical_keyword_name)
		end

	default_comparability: STRING is "-1"
			-- Default value for Daikon variable comparability

	modified_value: STRING is "1"
			-- Variable was modified (default value for the "modified bit")

	non_modified_value: STRING is "0"
			-- Variable was not modified			

	nonsensical_modified_value: STRING is "2"
			-- Variable can't be evaluated

	invocation_nonce_name: STRING is "this_invocation_nonce"

feature {NONE} -- Implementation
	old_suffixes: CI_OLD_SUFFIX_FORMAT is
			-- Old format for program point kinds
		once
			create Result
		end

	current_program_point: CI_PROGRAM_POINT
			-- Program point that is currently processed

	current_guard: ET_IF_INSTRUCTION
			-- If instruction that guards currently processed variable

	current_compound: ET_COMPOUND is
			-- Compound, where printing instruction should be inserted
		do
			if current_guard /= Void then
				Result := current_guard.then_compound
			else
				Result := state_printing_instructions
			end
		end

	alternative_compounds: LIST [ET_COMPOUND]
			-- Compounds, where alternative printing instruction should be inserted
			-- (when variable can't be evaluated, if its or any of its parents' precondition does not hold)	

	process_unfolded_variable (an_unfolded_variable: TREE [CI_VARIABLE]) is
			-- Process `an_unfolded_variable' in root - first child - ... - last child order
		require
			an_unfolded_variable_exists: an_unfolded_variable /= Void
		local
			old_guard: ET_IF_INSTRUCTION
			old_compound: ET_COMPOUND
			query: CI_QUERY
		do
			old_guard := current_guard
			if old_guard /= Void then
				old_compound := old_guard.then_compound
			else
				old_compound := state_printing_instructions
			end

			query ?= an_unfolded_variable.item
			if query /= Void and then query.relies_on_invariant and then not current_program_point.ensures_invariant then
				prepend_assertions (query.evaluation_preconditions, flat_invariants (query.target.type.base_class, query.type.universe))
			end
			if not an_unfolded_variable.item.evaluation_preconditions.is_empty then
				current_guard := guard (an_unfolded_variable.item.evaluation_preconditions)
				insert_instruction_to_back (old_compound, current_guard)
				alternative_compounds.extend (current_guard.else_compound)
			end

			process_variable (an_unfolded_variable.item)

			from
				an_unfolded_variable.child_start
			until
				an_unfolded_variable.child_after
			loop
				process_unfolded_variable (an_unfolded_variable.child)
				an_unfolded_variable.child_forth
			end

			if current_guard /= old_guard then
				alternative_compounds.finish
				alternative_compounds.remove
			end
			current_guard := old_guard
		end

	process_variable (a_variable: CI_VARIABLE) is
			-- Print static information about `a_variable'
			-- and add its printing instructions to `state_printing_instructions'
		require
			a_variable_exists: a_variable /= Void
		local
			synonim_local: ET_LOCAL_VARIABLE
			synonim_assignment: ET_ASSIGNMENT
		do
			output_variable_info (a_variable)

			if a_variable.constant = Void then
				-- Trace record should be output only for non-constant variables
				if a_variable.needs_synonim then
					create synonim_local.make (a_variable.synonim, a_variable.type.named_type)
					extra_locals.force_first (synonim_local)
					create synonim_assignment.make (a_variable.synonim, a_variable.as_expression)
					insert_instruction_to_back (current_compound, synonim_assignment)
				end

				insert_instruction_to_back (current_compound,
					variable_printing_instruction (a_variable.name, a_variable.identity, modified_value))

				from
					alternative_compounds.start
				until
					alternative_compounds.after
				loop
					insert_instruction_to_back (alternative_compounds.item,
						variable_printing_instruction (a_variable.name, nonsensical_keyword, nonsensical_modified_value))
					alternative_compounds.forth
				end
			end
		end

invariant
	current_compound_exists: current_compound /= Void
	declaration_keyword_not_empty: not declaration_keyword.is_empty
	nonsensical_keyword_exists: nonsensical_keyword /= Void
	default_comparability_not_empty: not default_comparability.is_empty
	default_modified_value_not_empty: not modified_value.is_empty
	nonsensical_modified_value_not_empty: not nonsensical_modified_value.is_empty
end
