indexing
	description: "Variables that are zero-argument queries on other variables"
	author: "Nadia Polikarpova"
	date: "$Date$"
	revision: "$Revision$"

class
	CI_QUERY

inherit
	CI_VARIABLE
		redefine
			eiffel_name,
			absolute_preconditions,
			needs_synonim,
			synonim_name,
			is_query,
--			constant,
			has_watched_once,
			is_same_writable,
			is_param
		end

	CI_OPERATOR_IDENTIFIERS
		export {NONE}
			all
		undefine
			is_equal
		end

	CI_RUNTIME_OUTPUT_METACLIENT
		export {NONE}
			all
		undefine
			is_equal
		end
create
	make

feature -- Initialization
	make (a_query: like ast_node; a_target: like target) is
			-- Create variable that corresponds to `a_query' on `a_target'
		require
			a_query_exist: a_query /= Void
			zero_arguments_in_a_query: a_query.arguments_count = 0
			a_target_exists: a_target /= Void
		local
			context: ET_TYPE_CONTEXT
			precondition: ET_EXPRESSION
			qualifier: CI_PRECONDITION_QUALIFIER
		do
			ast_node := a_query
			target := a_target
			surrounding_routine := target.surrounding_routine

			if target.is_current then
				context := target.type.context
			else
				context := target.type.context.new_type_context (target.type.base_type)
			end
			create type.make (ast_node.type, context, target.type.universe)

			clients := class_names_twin (target.clients)
			if not target.is_query then
				clients.force_first (target.type.surrounding_class.name)
			end

			create evaluation_preconditions.make
			if ast_node.is_function then
				precondition := flat_precondition (as_function, target.type.base_class, type.universe)
				if precondition /= Void then
					if not target.is_current then
						create qualifier.make (target.as_expression, surrounding_routine, context, type.universe)
						precondition := qualifier.qualified (precondition)
					end
					evaluation_preconditions.force_first (precondition)
				end
			end
			if Settings.is_watched_once (ast_node) then
				evaluation_preconditions.force_first (called_expression (ast_node.implementation_class.name, ast_node.name))
			end
			if not target.is_current and not target.type.kind.is_expanded then
				-- "Current" and expanded entities can't be void
				evaluation_preconditions.force_first (exists_condition (target.as_expression))
			end

			if needs_synonim then
				create synonim.make (synonim_name)
			end
		ensure
			ast_node_set: ast_node = a_query
			target_set: target = a_target
		end

feature -- Access
	target: CI_VARIABLE
			-- Target, on which query is called

	ast_node: ET_QUERY
			-- An AST node that corresponds to variable			

	name: STRING is
			-- String representation of query name in Daikon format
		do
			if not target.is_current then
				Result := target.name.string
				Result.append (".")
			else
				create Result.make_empty
			end
			Result.append (unqualified_name)
		end

	unqualified_name: STRING is
			-- String representation of query name in Daikon format without the target
		do
			Result := identifier_name (ast_node.name).lower_name
		end

	eiffel_name: STRING is
			-- String representation of query name in Eiffel format
		do
			if not target.is_current then
				Result := target.eiffel_name.string
				Result.append (".")
			else
				create Result.make_empty
			end
			Result.append (ast_node.name.name)
		end

--	constant: ET_CONSTANT is
--			-- Compile-time constant value of variable (void if not a constant)
--		local
--			const_attr: ET_CONSTANT_ATTRIBUTE
--		do
--			const_attr ?= ast_node
--			if const_attr /= Void and not type.is_string (type.base_class) then
--				Result := const_attr.constant
--			end
--		end

feature -- Evaluation
	as_expression: ET_CALL_EXPRESSION is
			-- Query call on target `a_target'
		do
			create Result.make (expression_target, ast_node.name, Void)
		end

	absolute_preconditions: ET_PRECONDITIONS is
			-- Absolute list of preconditions, which includes also `evaluation_preconditions' of all parent variables
		local
			parent_preconditions: ET_PRECONDITIONS
			i: INTEGER
		do
			parent_preconditions := target.absolute_preconditions
			create Result.make_with_capacity (parent_preconditions.count + evaluation_preconditions.count)
			from
				i := evaluation_preconditions.count
			until
				i < 1
			loop
				Result.put_first (evaluation_preconditions.item (i))
				i := i - 1
			end
			from
				i := parent_preconditions.count
			until
				i < 1
			loop
				Result.put_first (parent_preconditions.item (i))
				i := i - 1
			end
		end

feature {CI_VARIABLE_TYPE_KIND} -- Evaluation
	address_expression: ET_FEATURE_ADDRESS is
			-- Expression that yields address of this variable
		do
			if valid_as_address_target then
				create Result.make (ast_node.name)
			else
				create Result.make (synonim)
			end
		end

feature -- Status report
	needs_synonim: BOOLEAN is
			-- Does variable need a synonim or `as_expression' can be used everywhere as is?
		do
			Result := not valid_as_address_target and type.kind.uses_address
		end

	relies_on_invariant: BOOLEAN is
			-- Does query probably assume that class invariant holds on its entry?
		do
			Result := target.is_current and ast_node.is_function and not is_secret (ast_node)
		end

	has_watched_once: BOOLEAN is
			-- Does variable contain a watched one function?
		do
			Result := Settings.is_watched_once (ast_node) or target.has_watched_once
		end

	is_same_writable (a_value: ET_WRITABLE): BOOLEAN is
			-- Does this variable represent `a_writable'?
		local
			id: ET_IDENTIFIER
		do
			id ?= a_value
			Result := id /= Void and then target.is_current and id.same_call_name (ast_node.name)
		end

	is_param: BOOLEAN is
			-- Daikon property: variable value can't be changed by routine
		do
			Result := (surrounding_routine /= Void and then surrounding_routine.is_function) and target.is_param
		end

feature {CI_VARIABLE} -- Status report
	is_query: BOOLEAN is
			-- Is variable a query?
		once
			Result := True
		end

feature -- Suppression
	renamed (a_suppresion: CI_SUPPRESSION): CI_VARIABLE is
			-- Name of this variable in the suppressor program point from `a_suppression'
		do
			Result := a_suppresion.renamed_query (Current)
		end

feature {NONE} -- Implementation
	as_function: ET_FUNCTION is
			-- If `ast_node' is function: `ast_node'
		do
			Result ?= ast_node
		end

	expression_target: ET_EXPRESSION is
			-- Target to be used in `as_expression'
		do
			if not target.is_current then
				Result := target.as_expression
			end
		end

feature {CI_VARIABLE} -- Implementation		
	synonim_name: STRING is
			-- Name of `synonim'
		local
			str: STRING
		do
			str := identifier_name (ast_node.name).lower_name
			Result := target.synonim_name + str + "_"
		end

	valid_as_address_target: BOOLEAN is
			-- Can `as_expression' be used as target for address expression?
		do
			Result := target.is_current and ast_node.is_attribute and not ast_node.is_constant_attribute
		end

invariant
	ast_node_exists: ast_node /= Void
	zero_arguments: ast_node.arguments_count = 0
	target_exists: target /= Void
	as_function_exists_when_function: ast_node.is_function implies as_function /= Void
	unqualified_call: target.is_current implies expression_target = Void
	qualified_call: not target.is_current implies expression_target = target.as_expression
end
